@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2022 gnome-mpv
+ * Copyright (c) 2014-2022, 2024-2025 gnome-mpv
  *
  * This file is part of Celluloid.
  *
@@ -23,7 +23,9 @@
 #include <gio/gio.h>
 #include <gdk/gdk.h>
 #include <glib/gi18n.h>
+#include <adwaita.h>
 
+#include "celluloid-view.h"
 #include "celluloid-playlist-widget.h"
 #include "celluloid-playlist-model.h"
 #include "celluloid-playlist-item.h"
@@ -38,6 +40,9 @@ enum
 	PROP_0,
 	PROP_PLAYLIST_COUNT,
 	PROP_SEARCHING,
+	PROP_LOOP_FILE,
+	PROP_LOOP_PLAYLIST,
+	PROP_SHUFFLE,
 	N_PROPERTIES
 };
 
@@ -51,26 +56,42 @@ enum PlaylistColumn
 
 struct _CelluloidPlaylistWidget
 {
-	GtkBox parent_instance;
+	AdwBin parent_instance;
+	GSettings *settings;
 	gint64 playlist_count;
 	gboolean searching;
-	GtkListStore *store;
-	GtkWidget *scrolled_window;
 	CelluloidPlaylistModel *model;
-	GtkWidget *list_box;
 	gint last_selected;
 	gchar *drag_uri;
+	gint last_x;
+	gint last_y;
+	gboolean loop_file;
+	gboolean loop_playlist;
+	gboolean shuffle;
+	GtkCssProvider *css_provider;
 	GtkWidget *search_bar;
 	GtkWidget *search_entry;
 	GtkWidget *placeholder;
-	gint last_x;
-	gint last_y;
-	GtkCssProvider *css_provider;
+	GtkWidget *scrolled_window;
+	GtkWidget *list_box;
+	GtkWidget *toolbar_view;
+	GtkWidget *header_box;
+	GtkWidget *top_bar;
+	GtkWidget *bottom_bar;
+	GtkWidget *search_button;
+	GtkWidget *select_button;
+	GtkWidget *collapse_button;
+	GtkWidget *title_box;
+	GtkWidget *item_count;
+	GtkWidget *playlist_consecutive;
+	GtkWidget *loop1_btn;
+	GtkWidget *loop_list_btn;
+	GtkWidget *shuffle_btn;
 };
 
 struct _CelluloidPlaylistWidgetClass
 {
-	GtkBoxClass parent_class;
+	AdwBin parent_class;
 };
 
 static gint
@@ -87,6 +108,9 @@ find_match(	CelluloidPlaylistWidget *wgt,
 static void
 constructed(GObject *object);
 
+static void
+dispose(GObject *object);
+
 static void
 set_property(	GObject *object,
 		guint property_id,
@@ -169,7 +193,7 @@ drop_handler(	GtkDropTarget *self,
 static void
 realize_handler(GtkWidget *self, gpointer data);
 
-G_DEFINE_TYPE(CelluloidPlaylistWidget, celluloid_playlist_widget, GTK_TYPE_BOX)
+G_DEFINE_TYPE(CelluloidPlaylistWidget, celluloid_playlist_widget, ADW_TYPE_BIN)
 
 static gint
 get_selected_index(CelluloidPlaylistWidget *wgt)
@@ -241,48 +265,71 @@ find_match(	CelluloidPlaylistWidget *wgt,
 static GtkWidget *
 make_row(GObject *object, gpointer data)
 {
-	GtkWidget *row = gtk_list_box_row_new();
 	CelluloidPlaylistItem *item = CELLULOID_PLAYLIST_ITEM(object);
+	CelluloidPlaylistWidget *wgt = CELLULOID_PLAYLIST_WIDGET(data);
+	GtkWidget *row = gtk_list_box_row_new();
+	GtkWidget *row_vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 6);
+	GtkWidget *title_label = NULL;
+	GtkWidget *duration_label = NULL;
 	const gchar *title = celluloid_playlist_item_get_title(item);
 	const gchar *uri = celluloid_playlist_item_get_uri(item);
-	GtkWidget *label = NULL;
-	gchar *text = NULL;
+	const gint duration = (gint) celluloid_playlist_item_get_duration(item);
+	gchar *title_text = NULL;
+	gchar *duration_text = NULL;
 
 	if(title)
 	{
-		text = sanitize_utf8(title, TRUE);
+		title_text = sanitize_utf8(title, TRUE);
 	}
 	else
 	{
 		gchar *basename = g_path_get_basename(uri);
 
-		text = sanitize_utf8(basename, TRUE);
+		title_text = sanitize_utf8(basename, TRUE);
 		g_free(basename);
 	}
 
-	label = gtk_label_new(text);
-	g_free(text);
+	title_label = gtk_label_new(title_text);
+	gtk_widget_set_tooltip_text(title_label, title_text);
+	g_free(title_text);
 
-	g_assert(label);
+	g_assert(title_label);
 
 	if(celluloid_playlist_item_get_is_current(item))
 	{
-		PangoAttrList *attrs;
-		PangoAttribute *weight;
-
-		attrs = pango_attr_list_new();
-		weight = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
-
-		pango_attr_list_insert(attrs, weight);
-		gtk_label_set_attributes(GTK_LABEL(label), attrs);
-
-		pango_attr_list_unref(attrs);
+		gtk_widget_add_css_class(title_label, "heading");
 	}
 
-	gtk_widget_set_halign(label, GTK_ALIGN_START);
-	gtk_widget_set_margin_start(label, 6);
-	gtk_widget_set_margin_end(label, 6);
-	gtk_list_box_row_set_child(GTK_LIST_BOX_ROW(row), label);
+	gtk_widget_set_halign(title_label, GTK_ALIGN_START);
+	gtk_widget_set_margin_top(title_label, 12);
+	gtk_widget_set_hexpand(title_label, TRUE);
+	gtk_label_set_max_width_chars(GTK_LABEL(title_label), 40);
+	gtk_label_set_ellipsize(GTK_LABEL(title_label), PANGO_ELLIPSIZE_MIDDLE);
+
+	if(duration >= 0)
+	{
+		duration_text = format_time(duration, TRUE);
+	}
+
+	duration_label = gtk_label_new(duration_text);
+	g_free(duration_text);
+
+	g_assert(duration_label);
+
+	g_settings_bind(wgt->settings,
+			"show-durations-in-playlist",
+			duration_label,
+			"visible",
+			G_SETTINGS_BIND_GET );
+
+	gtk_widget_add_css_class(duration_label, "dim-label");
+	gtk_widget_set_halign(duration_label, GTK_ALIGN_START);
+	gtk_widget_set_margin_bottom(duration_label, 12);
+	gtk_label_set_ellipsize(GTK_LABEL(duration_label), PANGO_ELLIPSIZE_MIDDLE);
+
+	gtk_box_append(GTK_BOX(row_vbox), title_label);
+	gtk_box_append(GTK_BOX(row_vbox), duration_label);
+	gtk_list_box_row_set_child(GTK_LIST_BOX_ROW(row), row_vbox);
 
 	return row;
 }
@@ -292,27 +339,26 @@ constructed(GObject *object)
 {
 	CelluloidPlaylistWidget *self = CELLULOID_PLAYLIST_WIDGET(object);
 
-	self->store = gtk_list_store_new(	3,
-						G_TYPE_STRING,
-						G_TYPE_STRING,
-						G_TYPE_INT );
 	self->model = celluloid_playlist_model_new();
 	self->list_box = gtk_list_box_new();
 	self->last_selected = -1;
 	self->drag_uri = NULL;
+	self->toolbar_view = adw_toolbar_view_new();
 
 	gtk_list_box_set_selection_mode
 		(GTK_LIST_BOX(self->list_box), GTK_SELECTION_BROWSE);
 	gtk_list_box_set_activate_on_single_click
-		(GTK_LIST_BOX(self->list_box), FALSE);
+		(GTK_LIST_BOX(self->list_box), TRUE);
 	gtk_list_box_set_placeholder
 		(GTK_LIST_BOX(self->list_box), self->placeholder);
+	gtk_widget_add_css_class
+		(self->list_box, "navigation-sidebar");
 
 	gtk_list_box_bind_model
 		(	GTK_LIST_BOX(self->list_box),
 			G_LIST_MODEL(self->model),
 			(GtkListBoxCreateWidgetFunc)make_row,
-			NULL,
+			self,
 			NULL );
 
 	GtkGesture *click_gesture = gtk_gesture_click_new();
@@ -344,6 +390,7 @@ constructed(GObject *object)
 					G_N_ELEMENTS(types) );
 	gtk_widget_add_controller(	GTK_WIDGET(self->list_box),
 					GTK_EVENT_CONTROLLER(drop_target) );
+
 	g_signal_connect(	drop_target,
 				"motion",
 				G_CALLBACK(motion_handler),
@@ -410,18 +457,24 @@ constructed(GObject *object)
 	gtk_scrolled_window_set_child
 		(	GTK_SCROLLED_WINDOW(self->scrolled_window),
 			 GTK_WIDGET(viewport) );
-	gtk_box_append
-		(GTK_BOX(self), self->scrolled_window);
-
 	gtk_search_bar_set_child(GTK_SEARCH_BAR
 		(self->search_bar), self->search_entry);
-	gtk_box_append
-		(GTK_BOX(self), self->search_bar);
 
 	G_OBJECT_CLASS(celluloid_playlist_widget_parent_class)
 		->constructed(object);
 }
 
+static void
+dispose(GObject *object)
+{
+	CelluloidPlaylistWidget *self = CELLULOID_PLAYLIST_WIDGET(object);
+
+	g_clear_object(&self->settings);
+
+	G_OBJECT_CLASS(celluloid_playlist_widget_parent_class)
+		->dispose(object);
+}
+
 static void
 set_property(	GObject *object,
 		guint property_id,
@@ -429,23 +482,35 @@ set_property(	GObject *object,
 		GParamSpec *pspec )
 {
 	CelluloidPlaylistWidget *self = CELLULOID_PLAYLIST_WIDGET(object);
-
-	if(property_id == PROP_PLAYLIST_COUNT)
+	switch(property_id)
 	{
+		case PROP_PLAYLIST_COUNT:
 		self->playlist_count = g_value_get_int64(value);
-	}
-	else if(property_id == PROP_SEARCHING)
-	{
-		self->searching = g_value_get_boolean(value);
+		break;
 
+		case PROP_SEARCHING:
+		self->searching = g_value_get_boolean(value);
 		if(self->searching)
 		{
 			gtk_widget_grab_focus(self->search_entry);
 		}
-	}
-	else
-	{
+		break;
+
+		case PROP_LOOP_FILE:
+		self->loop_file = g_value_get_boolean(value);
+		break;
+
+		case PROP_LOOP_PLAYLIST:
+		self->loop_playlist = g_value_get_boolean(value);
+		break;
+
+		case PROP_SHUFFLE:
+		self->shuffle = g_value_get_boolean(value);
+		break;
+
+		default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
+		break;
 	}
 }
 
@@ -456,18 +521,31 @@ get_property(	GObject *object,
 		GParamSpec *pspec )
 {
 	CelluloidPlaylistWidget *self = CELLULOID_PLAYLIST_WIDGET(object);
-
-	if(property_id == PROP_PLAYLIST_COUNT)
+	switch(property_id)
 	{
+		case PROP_PLAYLIST_COUNT:
 		g_value_set_int64(value, self->playlist_count);
-	}
-	else if(property_id == PROP_SEARCHING)
-	{
+		break;
+
+		case PROP_SEARCHING:
 		g_value_set_boolean(value, self->searching);
-	}
-	else
-	{
+		break;
+
+		case PROP_LOOP_FILE:
+		g_value_set_boolean(value, self->loop_file);
+		break;
+
+		case PROP_LOOP_PLAYLIST:
+		g_value_set_boolean(value, self->loop_playlist);
+		break;
+
+		case PROP_SHUFFLE:
+		g_value_set_boolean(value, self->shuffle);
+		break;
+
+		default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
+		break;
 	}
 }
 
@@ -565,8 +643,8 @@ pressed_handler(	GtkGestureClick *gesture,
 				CELLULOID_MENU_ITEM(_("Add _Folder…"), "win.show-open-dialog((true, true))"),
 				CELLULOID_MENU_ITEM(_("Add _Location…"), "win.show-open-location-dialog(true)"),
 				CELLULOID_MENU_ITEM(_("_Shuffle"), "win.toggle-shuffle-playlist"),
-				CELLULOID_MENU_ITEM(_("Loop _File"), "win.toggle-loop-file"),
-				CELLULOID_MENU_ITEM(_("Loop _Playlist"), "win.toggle-loop-playlist"),
+				CELLULOID_MENU_ITEM(_("Loop File"), "win.toggle-loop-file"),
+				CELLULOID_MENU_ITEM(_("Loop Playlist"), "win.toggle-loop-playlist"),
 				CELLULOID_MENU_END };
 
 		gsize entries_offset = 0;
@@ -600,7 +678,7 @@ pressed_handler(	GtkGestureClick *gesture,
 		gtk_popover_set_has_arrow(GTK_POPOVER(ctx_menu), FALSE);
 		gtk_popover_set_pointing_to(GTK_POPOVER(ctx_menu), &rect);
 		gtk_widget_set_halign(ctx_menu, GTK_ALIGN_START);
-		gtk_widget_show(ctx_menu);
+		gtk_widget_set_visible(ctx_menu, TRUE);
 	}
 
 	return FALSE;
@@ -686,23 +764,34 @@ motion_handler(	GtkDropTarget *self,
 	{
 		GtkWidget *src = gtk_event_controller_get_widget(controller);
 		const gint row_h = gtk_widget_get_height(GTK_WIDGET(row));
-		gdouble row_x = 0;
-		gdouble row_y = 0;
 
-		gtk_widget_translate_coordinates
-			(src, GTK_WIDGET(row), x, y, &row_x, &row_y);
+		const graphene_point_t in_point =
+			{.x = (gfloat)x, .y = (gfloat)y};
+		graphene_point_t out_point =
+			{0};
 
-		const gboolean top_half = row_y < (row_h / 2);
-		gint row_index = gtk_list_box_row_get_index(row);
+		const gboolean computed =
+			gtk_widget_compute_point
+			(src, GTK_WIDGET(row), &in_point, &out_point);
 
-		css_data =
-			g_strdup_printf
-			(	"row:nth-child(%d)		"
-				"{				"
-				"	%s: 1px solid white;	"
-				"}				",
-				row_index + 1,
-				top_half ? "border-top" : "border-bottom" );
+		if(computed)
+		{
+			const gboolean top_half = out_point.y < (row_h / 2);
+			gint row_index = gtk_list_box_row_get_index(row);
+
+			css_data =
+				g_strdup_printf
+				(	"row:nth-child(%d)		"
+					"{				"
+					"	%s: 1px solid white;	"
+					"}				",
+					row_index + 1,
+					top_half ? "border-top" : "border-bottom" );
+		}
+		else
+		{
+			g_warning("Failed to calculate highlight position");
+		}
 	}
 	else
 	{
@@ -714,7 +803,7 @@ motion_handler(	GtkDropTarget *self,
 				"}					");
 	}
 
-	gtk_css_provider_load_from_data(wgt->css_provider, css_data, -1);
+	gtk_css_provider_load_from_string(wgt->css_provider, css_data);
 	g_free(css_data);
 
 	return GDK_ACTION_MOVE;
@@ -725,7 +814,7 @@ leave_handler(GtkDropTarget *self, gpointer data)
 {
 	CelluloidPlaylistWidget *wgt = CELLULOID_PLAYLIST_WIDGET(data);
 
-	gtk_css_provider_load_from_data(wgt->css_provider, "", -1);
+	gtk_css_provider_load_from_string(wgt->css_provider, "");
 }
 
 static gboolean
@@ -755,23 +844,33 @@ drop_handler(	GtkDropTarget *self,
 
 		if(dst_row)
 		{
-			gdouble row_x = 0;
-			gdouble row_y = 0;
+			const graphene_point_t in_point =
+				{.x = (gfloat)x, .y = (gfloat)y};
+			graphene_point_t out_point =
+				{0};
 
-			gtk_widget_translate_coordinates
-				(src, GTK_WIDGET(dst_row), x, y, &row_x, &row_y);
+			const gboolean computed =
+				gtk_widget_compute_point
+				(src, GTK_WIDGET(dst_row), &in_point, &out_point);
 
-			const gint row_h =
-				gtk_widget_get_height(GTK_WIDGET(dst_row));
-			const gboolean top_half =
-				row_y < (row_h / 2);
-			const gint dst_row_index =
-				gtk_list_box_row_get_index(dst_row);
-			const gint offset =
-				(top_half ? -1 : 0) +
-				(src_index - 1 < dst_row_index ? 0 : 1);
+			if(computed)
+			{
+				const gint row_h =
+					gtk_widget_get_height(GTK_WIDGET(dst_row));
+				const gboolean top_half =
+					out_point.y < (row_h / 2);
+				const gint dst_row_index =
+					gtk_list_box_row_get_index(dst_row);
+				const gint offset =
+					(top_half ? -1 : 0) +
+					(src_index - 1 < dst_row_index ? 0 : 1);
 
-			dst_index = CLAMP(dst_row_index + offset, 0, (gint)n_items - 1);
+				dst_index = CLAMP(dst_row_index + offset, 0, (gint)n_items - 1);
+			}
+			else
+			{
+				g_warning("Failed to calculate drop position");
+			}
 		}
 
 		GtkListBoxRow *new_dst_row =
@@ -779,7 +878,7 @@ drop_handler(	GtkDropTarget *self,
 			(GTK_LIST_BOX(wgt->list_box), dst_index);
 
 		gtk_list_box_select_row(GTK_LIST_BOX(wgt->list_box), new_dst_row);
-		gtk_css_provider_load_from_data(wgt->css_provider, "", -1);
+		gtk_css_provider_load_from_string(wgt->css_provider, "");
 
 		g_signal_emit_by_name
 			(	wgt,
@@ -800,7 +899,7 @@ drop_handler(	GtkDropTarget *self,
 
 			CelluloidPlaylistItem *item =
 				celluloid_playlist_item_new_take
-				(title, uri, FALSE);
+				(title, uri, 0, FALSE);
 
 			celluloid_playlist_model_append(wgt->model, item);
 			g_signal_emit_by_name(wgt, "row-inserted", position++);
@@ -811,7 +910,7 @@ drop_handler(	GtkDropTarget *self,
 		const gchar *string =
 			g_value_get_string(value);
 		CelluloidPlaylistItem *item =
-			celluloid_playlist_item_new(string, string, FALSE);
+			celluloid_playlist_item_new(string, string, 0, FALSE);
 
 		celluloid_playlist_model_append(wgt->model, item);
 		g_signal_emit_by_name(wgt, "row-inserted", n_items);
@@ -841,9 +940,12 @@ celluloid_playlist_widget_class_init(CelluloidPlaylistWidgetClass *klass)
 	GParamSpec *pspec = NULL;
 
 	obj_class->constructed = constructed;
+	obj_class->dispose = dispose;
 	obj_class->set_property = set_property;
 	obj_class->get_property = get_property;
 
+	gtk_widget_class_set_css_name(GTK_WIDGET_CLASS(klass), "playlist");
+
 	pspec = g_param_spec_int64
 		(	"playlist-count",
 			"Playlist count",
@@ -862,6 +964,30 @@ celluloid_playlist_widget_class_init(CelluloidPlaylistWidgetClass *klass)
 			G_PARAM_READWRITE );
 	g_object_class_install_property(obj_class, PROP_SEARCHING, pspec);
 
+	pspec = g_param_spec_boolean
+		(	"loop-file",
+			"Loop file",
+			"Loop a single media file",
+			FALSE,
+			G_PARAM_READWRITE );
+	g_object_class_install_property(obj_class, PROP_LOOP_FILE, pspec);
+
+	pspec = g_param_spec_boolean
+		(	"loop-playlist",
+			"Loop playlist",
+			"Loop current playlist",
+			FALSE,
+			G_PARAM_READWRITE );
+	g_object_class_install_property(obj_class, PROP_LOOP_PLAYLIST, pspec);
+
+	pspec = g_param_spec_boolean
+		(	"shuffle",
+			"Shuffle playlist",
+			"Shuffle current playlist",
+			FALSE,
+			G_PARAM_READWRITE );
+	g_object_class_install_property(obj_class, PROP_SHUFFLE, pspec);
+
 	g_signal_new(	"row-activated",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST,
@@ -908,25 +1034,150 @@ celluloid_playlist_widget_class_init(CelluloidPlaylistWidgetClass *klass)
 static void
 celluloid_playlist_widget_init(CelluloidPlaylistWidget *wgt)
 {
-	GtkStyleContext *style = NULL;
-
+	wgt->settings = g_settings_new(CONFIG_ROOT);
 	wgt->playlist_count = 0;
 	wgt->searching = FALSE;
+	wgt->loop_file = FALSE;
+	wgt->loop_playlist = FALSE;
+	wgt->shuffle = FALSE;
 	wgt->scrolled_window = gtk_scrolled_window_new();
 	wgt->search_bar = gtk_search_bar_new();
 	wgt->search_entry = gtk_search_entry_new();
-	wgt->placeholder = gtk_label_new(_("Playlist is empty"));
+	wgt->placeholder = adw_status_page_new();
 	wgt->css_provider = gtk_css_provider_new();
+	wgt->toolbar_view = adw_toolbar_view_new();
+	wgt->top_bar = adw_header_bar_new();
+	wgt->collapse_button = gtk_button_new_from_icon_name("go-next-symbolic");
+	wgt->title_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 2);
+	wgt->search_button = gtk_toggle_button_new();
+	wgt->bottom_bar = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
+	wgt->playlist_consecutive = gtk_toggle_button_new();
+	wgt->item_count = gtk_label_new(_("0 files"));
+	wgt->loop1_btn = gtk_toggle_button_new();
 
-	style = gtk_widget_get_style_context(wgt->placeholder);
-	gtk_style_context_add_class(style, "dim-label");
+	adw_header_bar_set_show_end_title_buttons
+		(ADW_HEADER_BAR(wgt->top_bar), FALSE);
+	adw_header_bar_pack_start
+		(ADW_HEADER_BAR(wgt->top_bar), wgt->collapse_button);
 
-	gtk_widget_set_vexpand
-		(wgt->scrolled_window, TRUE);
-	gtk_label_set_ellipsize
-		(GTK_LABEL(wgt->placeholder), PANGO_ELLIPSIZE_END);
-	gtk_orientable_set_orientation
-		(GTK_ORIENTABLE(wgt), GTK_ORIENTATION_VERTICAL);
+	gtk_actionable_set_action_name
+		(GTK_ACTIONABLE(wgt->collapse_button), "win.toggle-playlist");
+	gtk_widget_set_tooltip_text
+		(wgt->collapse_button, _("Hide playlist"));
+
+	gtk_box_append
+		(GTK_BOX(wgt->title_box), gtk_label_new(_("Playlist")));
+	gtk_widget_add_css_class
+		(gtk_widget_get_first_child(wgt->title_box), "heading");
+
+	gtk_widget_set_halign(wgt->item_count, GTK_ALIGN_START);
+	gtk_box_append(GTK_BOX(wgt->title_box), wgt->item_count);
+	gtk_widget_add_css_class(wgt->item_count, "caption");
+
+	adw_header_bar_pack_start
+		(ADW_HEADER_BAR(wgt->top_bar), wgt->title_box);
+	adw_header_bar_pack_end
+		(ADW_HEADER_BAR(wgt->top_bar), wgt->search_button);
+	adw_header_bar_set_show_title
+		(ADW_HEADER_BAR(wgt->top_bar), FALSE);
+
+	gtk_widget_set_margin_top(wgt->top_bar, 6);
+	gtk_widget_set_margin_bottom(wgt->top_bar, 6);
+	gtk_widget_set_margin_start(wgt->top_bar, 6);
+	gtk_widget_set_margin_end(wgt->top_bar, 6);
+
+	gtk_button_set_icon_name
+		(GTK_BUTTON(wgt->search_button), "system-search");
+	g_object_bind_property
+		(	wgt->search_bar, "search-mode-enabled",
+			wgt->search_button, "active",
+			G_BINDING_BIDIRECTIONAL );
+	gtk_widget_set_tooltip_text
+		(wgt->search_button, _("Search in playlist"));
+
+	gtk_search_entry_set_placeholder_text
+		(GTK_SEARCH_ENTRY(wgt->search_entry), _("Search in playlist…"));
+
+	gtk_widget_set_halign(wgt->bottom_bar, GTK_ALIGN_CENTER);
+	gtk_widget_add_css_class(wgt->bottom_bar, "toolbar");
+
+	gtk_button_set_icon_name
+		(	GTK_BUTTON(wgt->playlist_consecutive),
+			"media-playlist-consecutive-symbolic" );
+
+	gtk_box_append(GTK_BOX(wgt->bottom_bar), wgt->playlist_consecutive);
+
+	gtk_toggle_button_set_active
+		(GTK_TOGGLE_BUTTON(wgt->playlist_consecutive),TRUE);
+	gtk_widget_set_tooltip_text
+		(wgt->playlist_consecutive, _("Disable repeat"));
+
+	gtk_button_set_icon_name
+		(GTK_BUTTON(wgt->loop1_btn), "media-playlist-repeat-song-symbolic");
+	gtk_box_append
+		(GTK_BOX(wgt->bottom_bar), wgt->loop1_btn);
+	gtk_toggle_button_set_group
+		(	GTK_TOGGLE_BUTTON(wgt->loop1_btn),
+			GTK_TOGGLE_BUTTON(wgt->playlist_consecutive) );
+	gtk_widget_set_tooltip_text
+		(wgt->loop1_btn, _("Repeat file"));
+
+	wgt->loop_list_btn = gtk_toggle_button_new();
+	gtk_button_set_icon_name
+		(GTK_BUTTON(wgt->loop_list_btn), "media-playlist-repeat-symbolic");
+	gtk_box_append
+		(GTK_BOX(wgt->bottom_bar), wgt->loop_list_btn);
+	gtk_toggle_button_set_group
+		(	GTK_TOGGLE_BUTTON(wgt->loop_list_btn),
+			GTK_TOGGLE_BUTTON(wgt->playlist_consecutive) );
+	gtk_widget_set_tooltip_text
+		(wgt->loop_list_btn, _("Repeat playlist"));
+
+	wgt->shuffle_btn = gtk_toggle_button_new();
+	gtk_button_set_icon_name
+		(GTK_BUTTON(wgt->shuffle_btn), "media-playlist-shuffle-symbolic");
+	gtk_box_append
+		(GTK_BOX(wgt->bottom_bar), wgt->shuffle_btn);
+	gtk_widget_set_tooltip_text
+		(wgt->shuffle_btn, _("Shuffle playlist"));
+
+	wgt->scrolled_window = gtk_scrolled_window_new();
+	wgt->header_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
+	gtk_box_append(GTK_BOX(wgt->header_box), wgt->top_bar);
+	gtk_box_append(GTK_BOX(wgt->header_box), wgt->search_bar);
+
+	adw_bin_set_child
+		(ADW_BIN(wgt), wgt->toolbar_view);
+	adw_toolbar_view_add_top_bar
+		(ADW_TOOLBAR_VIEW(wgt->toolbar_view), wgt->header_box);
+	adw_toolbar_view_add_bottom_bar
+		(ADW_TOOLBAR_VIEW(wgt->toolbar_view), wgt->bottom_bar);
+	adw_toolbar_view_set_content
+		(ADW_TOOLBAR_VIEW(wgt->toolbar_view), wgt->scrolled_window);
+
+	gchar *css_data =
+		"playlist .icon { -gtk-icon-size: 64px; }\n"
+		"playlist .title { font-weight: normal; font-size: medium; }\n";
+
+	gtk_css_provider_load_from_string(wgt->css_provider, css_data);
+	gtk_widget_add_css_class(wgt->placeholder, "dim-label");
+
+	adw_status_page_set_title
+		(	ADW_STATUS_PAGE(wgt->placeholder),
+			_("Playlist is Empty") );
+	adw_status_page_set_icon_name
+		(	ADW_STATUS_PAGE(wgt->placeholder),
+			"applications-multimedia-symbolic" );
+
+	g_object_bind_property(	wgt, "loop-file",
+				wgt->loop1_btn, "active",
+				G_BINDING_BIDIRECTIONAL );
+	g_object_bind_property(	wgt, "loop-playlist",
+				wgt->loop_list_btn, "active",
+				G_BINDING_BIDIRECTIONAL );
+	g_object_bind_property(	wgt, "shuffle",
+				wgt->shuffle_btn, "active",
+				G_BINDING_BIDIRECTIONAL );
 }
 
 GtkWidget *
@@ -1002,6 +1253,7 @@ celluloid_playlist_widget_update_contents(	CelluloidPlaylistWidget *wgt,
 			celluloid_playlist_item_new_take
 			(	g_strdup(entry->title),
 				g_strdup(entry->filename),
+				entry->duration,
 				(gint)i == current	);
 
 		celluloid_playlist_model_append(wgt->model, item);
@@ -1014,6 +1266,17 @@ celluloid_playlist_widget_update_contents(	CelluloidPlaylistWidget *wgt,
 
 	wgt->playlist_count = playlist->len;
 	g_object_notify(G_OBJECT(wgt), "playlist-count");
+
+	const gulong playlist_count =
+		(gulong)wgt->playlist_count;
+	gchar *label_text =
+		g_strdup_printf
+		(	ngettext("%ld file", "%ld files", playlist_count),
+			playlist_count );
+
+	gtk_label_set_label(GTK_LABEL(wgt->item_count), label_text);
+
+	g_free(label_text);
 }
 
 GPtrArray *
